import json
import numbers
import pprint
import requests
import warnings

from requests.exceptions import ConnectionError

from ..util import add_apikey_header, AlertMixin
from .query import Metadata, ApiFilters

class PostException(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)

class MetadataPostException(PostException): pass
class EventTypePostException(PostException): pass
class EventTypeBulkPostException(PostException): pass

class MetadataPostWarning(Warning): pass
class EventTypePostWarning(Warning): pass
class EventTypeBulkPostWarning(Warning): pass


class PostBase(AlertMixin, object):
    """
    Base class for perfsonar API post functionality.  Should not 
    be directly instantiated.
    """
    _schema_root = 'perfsonar/archive'
    def __init__(self, api_url, username, api_key, script_alias):
        """
        The api_url, username and api_key args all have their usual usages.

        The script_alias arg (which defaults to 'esmond') is in place
        for the perfsonar CentOS distributions.  The 'root' of the 
        perfsonar api as defined in esmond.api.perfsonar.api is 
        is http://host:port/perfsonar/archive/... When deployed under 
        Apache on the canned perfsonar installations, a ScriptAlias 
        of /esmond is employed so the rest api won't 'take over' the 
        default webserver.  When running this way it yields a base 
        of http://host:port/esmond/perfsonar/archive/... - this is 
        the default of the canned system install.

        The script_alias arg can be set to None when doing development
        against the django dev runserver, or set to something else if 
        running under a similar but different deployment.  

        Setting the arg to '/' will yield the same result as setting it 
        to None - this makes it easier for calling programs that are 
        setting that value via a command line arg/config file/etc.
        """
        super(PostBase, self).__init__()
        self.api_url = api_url
        if self.api_url: self.api_url = api_url.rstrip('/')

        self.username = username
        self.api_key = api_key
        self.script_alias = script_alias
        self.headers = { 'content-type': 'application/json' }
        if self.username and self.api_key:
            add_apikey_header(self.username, self.api_key, self.headers)
        else:
            self.warn('username and api_key not set')
        # sublass override
        self._payload = {}

        try:
            getattr(self, 'wrn')
        except AttributeError:
            raise PostException('Do not instantiate base class, use appropriate subclass')

        if self.script_alias: 
            self.script_alias = script_alias.rstrip('/')
            self.script_alias = script_alias.lstrip('/')
            if self.script_alias:
                self._schema_root = '{0}/{1}'.format(self.script_alias, self._schema_root)

    def _validate(self):
        """Will be overridden in subclass.  Needs to run whatever validation
        checks on the internal payload data before sending it to the API."""
        raise NotImplementedError('Must be implemented in subclass')

    def json_payload(self, pp=False):
        self._validate()
        if pp:
            return json.dumps(self._payload, indent=4)
        else:
            return json.dumps(self._payload)
        
class MetadataPost(PostBase):
    wrn = MetadataPostWarning
    """
    Client class to POST metadata data/objects to the perfsonar 
    relational database.
    """
    def __init__(self, api_url, username='', api_key='',
            subject_type=None, source=None, destination=None,
            tool_name=None, measurement_agent=None, input_source=None,
            input_destination=None, time_duration=None, 
            ip_transport_protocol=None, script_alias='esmond'):
        super(MetadataPost, self).__init__(api_url, username, api_key, script_alias)
        # required
        self.subject_type = subject_type
        self.source = source
        self.destination = destination
        self.tool_name = tool_name
        self.measurement_agent = measurement_agent
        self.input_source = input_source
        self.input_destination = input_destination
        # optional - if not defined, don't send at all
        self.time_duration = time_duration
        self.ip_transport_protocol = ip_transport_protocol

        self._required_args = (
            'subject_type',
            'source',
            'destination',
            'tool_name',
            'measurement_agent',
            'input_source',
            'input_destination'
        )

        self._optional_args = (
            'time_duration', 
            'ip_transport_protocol'
        )

        self._payload = {
            'subject-type': self.subject_type,
            'source': self.source,
            'destination': self.destination,
            'tool-name': self.tool_name,
            'measurement-agent': self.measurement_agent,
            'input-source': self.input_source,
            'input-destination': self.input_destination,
            'time-duration': self.time_duration,
            'ip-transport-protocol': self.ip_transport_protocol,
            'event-types': []
        }

    def add_event_type(self, et):
        """Add event-type data to the metadata before POSTing to 
        the backend.  Will normally be called more than once."""

        for i in self._payload['event-types']:
            if i.get('event-type', None) == et:
                self.warn('Event type {0} already exists - skipping'.format(et))

        self._payload['event-types'].append({ 'event-type' : et, 'summaries': []})

    def add_summary_type(self, et, st, windows=[]):
        """Add associated summary-type data to metadata before POSTing."""

        if not windows:
            self.warn('No summary windows were defined - skipping')
            return

        for i in windows:
            try:
                int(i)
            except ValueError:
                raise MetadataPostException('Invalid summary window int: {0}'.format(i))

        for existing_et_def in self._payload['event-types']:
            if existing_et_def.get('event-type', None) == et:
                #add new summaries
                new_summaries = [{ 'summary-type': st, 'summary-window': x } for x in windows]
                #clear out old summaries
                for summ in existing_et_def['summaries']:
                    if (summ['summary-type'] == st):
                        #remove any existing summaries of same type
                        pass
                    else:
                         new_summaries.append(summ)
                existing_et_def['summaries'] = new_summaries
                return

        
        #if event type does not exist then create and add summaries
        suminfo = {
            'event-type': et,
            'summaries' : [ 
                { 'summary-type': st, 'summary-window': x } for x in windows
            ]
        }

        self._payload['event-types'].append(suminfo)

    def add_freeform_key_value(self, k, v):
        """Add key/values pairs to metadata payload to be stored in the
        ps_metadata_parameters table."""
        if not self._payload.get(k, None):
            self._payload[k] = v
        else:
            self.warn('Payload key {0} exists - skipping'.format(k))

    def post_metadata(self):
        """Void method that will post the new metadata to the API and 
        return the newly created metadata information wrapped in an
        esmond.api.perfsonar.query.Metadata object."""
        url = '{0}/{1}/'.format(self.api_url, self._schema_root)

        try:
            r = requests.post(url, data=self.json_payload(), headers=self.headers)
        except ConnectionError, e:
            self.warn('POST connection error: {0}'.format(str(e)))
            return None

        if not r.status_code == 201:
            # Change this to an exception?
            self.warn('POST error: status_code: {0}, message: {1}'.format(r.status_code, r.content))
            return None

        return Metadata(json.loads(r.content), self.api_url, ApiFilters())

    def _validate(self):
        redash = lambda s: s.replace('_', '-')

        # make sure required args are present
        for arg in self._required_args:
            if not getattr(self, arg):
                raise MetadataPostException('Reqired arg {0} not set'.format(arg))

        # remove any optional args from payload
        for arg in self._optional_args:
            if not getattr(self, arg) and \
                self._payload.has_key(redash(arg)):
                del self._payload[redash(arg)]


class EventTypePost(PostBase):
    wrn = EventTypePostWarning
    """Client code to POST data to a single event type to the perfsonar API."""
    def __init__(self, api_url, username='', api_key='', metadata_key=None,
            event_type=None, script_alias='esmond'):
        """
        The api_url, username and api_key args have their usual uses.  The
        event_type arg is outlined in the base class.

        The metadata_key and event_type args are the associated metadata 
        / event-type new data is being added to.
        """
        super(EventTypePost, self).__init__(api_url, username, api_key, script_alias)

        self.metadata_key = metadata_key
        self.event_type = event_type

        if not self.metadata_key or not self.event_type:
            raise EventTypePostException('Must set metadata_key and event_type args')

        # Slightly different than payload in meta data class.
        # List will hold data points, but whole thing will not
        # be sent to server.
        self._payload = []

    def add_data_point(self, ts, val):
        """Add a new ts/datapoint to the payload for the meta/event type 
        before sending to api.  Will be called more than once."""
        if not isinstance(ts, int):
            raise EventTypePostException('ts arg must be an integer')
        self._payload.append( { 'ts': ts, 'val': val } )

    def post_data(self):
        """Void method to send the payload to the API.  Does not return 
        anything."""
        self._validate()

        url = '{0}/{1}/{2}/{3}/base'.format(self.api_url, self._schema_root,
            self.metadata_key, self.event_type)

        results = []

        for dp in self._payload:
            try:
                r = requests.post(url, data=json.dumps(dp), headers=self.headers)
            except ConnectionError, e:
                self.warn('POST connection error: {0}'.format(str(e)))

            if not r.status_code == 201:
                self.warn('POST error: status_code: {0}, message: {1}'.format(r.status_code, r.content))

    def _validate(self):
        for i in self._payload:
            if isinstance(i['val'], dict):
                for k,v in i['val'].items():
                    try:
                        int(k), int(v)
                    except ValueError:
                        raise EventTypePostException('Histogram dict items must be integer values - got {0} {1}'.format(k,v))

class EventTypeBulkPost(PostBase):
    wrn = EventTypePostWarning
    """Client class to bulk post data.  This is used to write multiple 
    event types to a single metadata."""
    def __init__(self, api_url, username='', api_key='', metadata_key=None, 
            script_alias='esmond'):
        """
        The api_url, username and api_key args have their usual uses.  The
        event_type arg is outlined in the base class.  The metadata_key arg
        is the string for the metadata being written to.
        """
        super(EventTypeBulkPost, self).__init__(api_url, username, api_key, script_alias)

        self.metadata_key = metadata_key

        if not self.metadata_key:
            raise EventTypeBulkPostException('Must set metadata_key')

        self._payload = { 'data': [] }

    def _get_ts_payload_entry(self, ts):
        entry = False

        for i in self._payload['data']:
            if i['ts'] == ts: entry = i

        if not entry:
            self._payload['data'].append( {'ts': ts, 'val': [] } )
            return self._get_ts_payload_entry(ts)

        return entry

    def add_data_point(self, event_type, ts, val):
        """Adds data points to the payload.  Adds new ts/datapoint to
        a specific event-type in the internal payload.  Will be called
        multiple times."""

        if not isinstance(ts, int):
            raise EventTypeBulkPostException('ts arg must be an integer')

        data_entry = self._get_ts_payload_entry(ts)
        data_entry['val'].append({'event-type': event_type, 'val': val})

    def post_data(self):
        """Void method to send the payload to the API.  Does not return 
        anything."""
        self._validate()

        url = '{0}/{1}/{2}/'.format(self.api_url, self._schema_root,
            self.metadata_key)
        
        try:
            r = requests.post(url, data=self.json_payload(), headers=self.headers)
        except ConnectionError, e:
            self.warn('POST connection error: {0}'.format(str(e)))

        if not r.status_code == 201:
            # Change this to an exception?
            self.warn('POST error: status_code: {0}, message: {1}'.format(r.status_code, r.content))

    def _validate(self):
        for i in self._payload['data']:
            # Validation currently being done on incoming values
            # in add_data_point - leaving this here for the future.
            pass



        




import json
import requests
import warnings

from .snmp import DataPayload
from .util import add_apikey_header, atencode, AlertMixin

class TimeseriesBase(AlertMixin, object):
    """Base class for the GET and POST timeseries interaction objects."""
    _schema_root = 'v1/timeseries'
    def __init__(self, api_url='http://localhost/', path=[], freq=None,
        username='', api_key=''):
        """Constructor - the path list arg is an ordered list of elements
        that will be used (along with the freq arg) to construct the 
        cassandra row key.  See example above."""
        super(TimeseriesBase, self).__init__()
        self.api_url = api_url.rstrip("/")
        self.path = path[:] # copy in case the path ref is reused
        self.freq = freq
        self.username = username
        self.api_key = api_key

        # Error from base classes
        self.errno = None
        self.errms = None

        # Make sure we're not using the base class
        try:
            getattr(self, '_p_type')
        except AttributeError:
            raise PostException('Do not instantiate TimeseriesBase base class, use appropriate subclass.')

        # Validate args
        if not self.path or not self.freq or not self.api_url:
            raise PostException('The args api_url, path and freq must be set.')

        if not isinstance(self.path, list):
            raise PostException('Path argument must be a list.')

        if not len(self.path) > 1:
            raise PostException('Path is not of sufficient length.')

        try:
            int(self.freq)
        except ValueError:
            raise PostException('Arg freq must be an integer.')

        self.path = [ atencode(step) for step in self.path ]

        self.url = '{0}/{1}/{2}/{3}/{4}'.format(self.api_url, 
            self._schema_root, self._p_type,
            '/'.join(self.path), self.freq)

    def set_error_state(self, errno, errms):
        self.errno = errno
        self.errms = errms

    @property
    def get_error(self):
        if self.errno or self.errms:
            return '{0}: {1}'.format(self.errno, self.errms)
        return None



"""
Classes to handle posting data to esmond rest interface.

Example use:

    ts = int(time.time()) * 1000

    payload = [
        { 'ts': ts-90000, 'val': 1000 },
        { 'ts': ts-60000, 'val': 2000 },
        { 'ts': ts-30000, 'val': 3000 },
        { 'ts': ts, 'val': 4000 },
    ]

    path = ['rtr_test_post', 'FastPollHC', 'ifHCInOctets', 'interface_test']

    p = PostRawData(port=8000, path=path, freq=30000)
    p.set_payload(payload)
    p.add_to_payload({'ts': ts+1000, 'val': 5000})
    p.send_data()

The internal payload is initialized to an empty list.  The client can either
compltely overwrite the internal payload with the set_payload() method, or 
one can add indivdual dicts to it with add_to_payload().  

In either case, the payload will be internally validated and an exception
will be raised if bad data is found.
"""

class PostWarning(Warning): pass
class PostRawDataWarning(PostWarning): pass
class PostBaseRateWarning(PostWarning): pass

class PostException(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)

class PostData(TimeseriesBase):
    """Base class for API write objects - writes data to the POST 
    facility in the /timeseries/ REST interface namespace."""
    wrn = PostWarning

    def __init__(self, api_url='http://localhost/', path=[], freq=None,
        username='', api_key=''):
        """Constructor - the path list arg is an ordered list of elements
        that will be used (along with the freq arg) to construct the 
        cassandra row key.  See example above."""
        super(PostData, self).__init__(api_url, path, freq, username, api_key)

        # Make sure we're not using an intermediate base class.
        try:
            getattr(self, '_p_type')
        except AttributeError:
            raise PostException('Do not instantiate PostData base class, use appropriate subclass.')

        if not self.username or not self.api_key:
            raise PostException('PostData requires username and api_key for rest interface.')

        # Set up playload and headers.

        self.payload = []
        self.headers = { 'content-type': 'application/json' }
        add_apikey_header(self.username, self.api_key, self.headers)

    def set_payload(self, payload):
        """Sets object payload to a complete list of dicts passed in. 
        This will overwrite the internal payload if any elements had 
        been previously defined (and issue a warning)."""
        if not isinstance(payload, list):
            raise PostException('Arg payload to set_payload must be a list instance.')

        if len(self.payload):
            self.warn('Internal payload was not empty, so this is a warning that you are overwriting an existing payload.')

        self.payload = payload

        self._validate_payload()

    def add_to_payload(self, item):
        """Adds a new dict element to the object payload."""
        if not isinstance(item, dict):
            raise PostException('Arg item to add_to_payload must be a dict instance.')

        self.payload.append(item)

        self._validate_payload()

    def _validate_payload(self):
        """Internal method to check the integrity of the payload whenever 
        it is set or appended to."""
        for i in self.payload:
            if not isinstance(i, dict):
                raise PostException('All elements of payload must be dicts - got: {0}'.format(i))
            if not i.has_key('ts') or not i.has_key('val'):
                raise PostException('Expecting list of dicts with keys \'val\' and \'ts\' - got: {0}'.format(i))
            try:
                int(float(i.get('ts')))
                float(i.get('val'))
            except ValueError:
                raise PostException('Must supply valid numeric args for ts and val dict attributes - got: {0}'.format(i))

    def send_data(self):
        """Format current payload, send to REST api and clear the payload.
        Payload is cleared after a write so as not to send duplicate data, 
        and so the same instance can be used to send multiple times."""

        if not self.payload:
            self.warn('Payload empty, no data sent.')
            return

        r = requests.post(self.url, data=json.dumps(self.payload), headers=self.headers)

        if not r.status_code == 201:
            # Change this to an exception?
            self.set_error_state(r.status_code, r.content)
            self.warn('POST error: status_code: {0}, message: {1}'.format(r.status_code, r.content))

        # reset payload
        self.payload = []



class PostRawData(PostData):
    """Class to post raw data to rest api."""
    _p_type = 'RawData'
    wrn = PostRawDataWarning

class PostBaseRate(PostData):
    """Class to post base rate deltas to rest api."""
    _p_type = 'BaseRate'
    wrn = PostBaseRateWarning

"""
Classes to execute get requests to timeseries namespace.

These classes are initialized much like the post data class, but with 
and additional params dict arg that will be passed to the GET request.
Setting begin or end time would be the most common args set.  Params dict 
is optional.

Get data will return a payload encapsulation object which will return 
data point objects.  Since this namespace is supposed to just return 
that which is in the backend - the timestamps are all in milliseconds 
and there is no coersion to a python datetime object.

Example usage:

    path = ['rtr_test_post', 'FastPollHC', 'ifHCInOctets', 'interface_test/0/0.0']

    params = {
        'begin': ts-90000, 'end': ts+1000
    }

    get = None

    args = {
        'api_url': 'http://localhost:8000/', 
        'path': path, 
        'freq': 30000,
        'params': params,
    }

    if p_type == 'RawData':
        get = GetRawData(**args)
    elif p_type == 'BaseRate':
        get = GetBaseRate(**args)

    payload = get.get_data()

    print payload
    for d in payload.data:
        print '  *', d
"""

class GetWarning(Warning): pass
class GetRawDataWarning(GetWarning): pass
class GetBaseRateWarning(GetWarning): pass

class GetException(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)

class GetData(TimeseriesBase):
    """Base class for API write objects - writes data to the Get 
    facility in the /timeseries/ REST interface namespace."""
    wrn = GetWarning
    def __init__(self, api_url='http://localhost/', path=[], freq=None, 
        username='', api_key='', params={}):
        """Constructor - the path list arg is an ordered list of elements
        that will be used (along with the freq arg) to construct the 
        cassandra row key.  See example above."""
        super(GetData, self).__init__(api_url, path, freq, username, api_key)

        self._params = params

        # Make sure we're not using an intermediate base class.
        try:
            getattr(self, '_p_type')
        except AttributeError:
            raise PostException('Do not instantiate GetData base class, use appropriate subclass.')

        self.headers = {}
        if self.username and self.api_key:
            add_apikey_header(self.username, self.api_key, self.headers)

    def get_data(self):
        r = requests.get(self.url, params=self._params, headers=self.headers)
        if r.status_code == 200 and \
            r.headers['content-type'] == 'application/json':
            data = json.loads(r.content)
            return TimeSeriesDataPayload(data)
        else:
            self.set_error_state(r.status_code, r.content)
            self.warn('GET error: status_code: {0}, message: {1}'.format(r.status_code, r.content))
            return TimeSeriesDataPayload()

class GetRawData(GetData):
    """Class to Get raw data to rest api."""
    _p_type = 'RawData'
    wrn = GetRawDataWarning

class GetBaseRate(GetData):
    """Class to Get base rate deltas to rest api."""
    _p_type = 'BaseRate'
    wrn = GetBaseRateWarning

"""
Classes to make bulk data requests.
"""

class GetBulkWarning(Warning): pass
class GetBulkRawDataWarning(GetWarning): pass
class GetBulkBaseRateWarning(GetWarning): pass

class BulkException(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)

class GetBulkData(AlertMixin, object):
    wrn = GetBulkWarning
    _schema_root = 'v1/bulk/timeseries'
    def __init__(self, api_url='http://localhost', username='', api_key=''):
        super(GetBulkData, self).__init__()
        self.api_url = api_url.rstrip("/")
        self.username = username
        self.api_key = api_key

        # Make sure we're not using the base class
        try:
            getattr(self, '_p_type')
        except AttributeError:
            raise BulkException('Do not instantiate GetBulkData base class, use appropriate subclass.')

        # Validate args
        if not self.api_url:
            raise BulkException('The arg api_url must be set.')

        self.headers = { 'content-type': 'application/json' }
        if self.username and self.api_key:
            add_apikey_header(self.username, self.api_key, self.headers)
        self.url = '{0}/{1}/'.format(self.api_url, self._schema_root)

    def get_data(self, paths, begin=None, end=None):

        self._validate_paths(paths)
        self._validate_args(begin=begin, end=end)

        payload = {
            'type': self._p_type,
            'paths': paths,
        }

        if begin: payload['begin'] = begin
        if end: payload['end'] = end

        r = requests.post(self.url, data=json.dumps(payload), headers=self.headers)

        if r.status_code == 201 and \
            r.headers['content-type'] == 'application/json':
            data = json.loads(r.content)
            return TimeSeriesBulkDataPayload(data)
        else:
            self.warn('GET error: status_code: {0}, message: {1}'.format(r.status_code, r.content))
            return TimeSeriesBulkDataPayload()

    def _validate_paths(self, paths):
        if not isinstance(paths, list):
            raise BulkException('Paths argument must be a list.')

        if not len(paths):
            raise BulkException('Paths list contains no elements.')

        for path in paths:
            if not path:
                raise BulkException('Path elements must contain data.')
            try:
                int(path[-1])
            except ValueError:
                raise BulkException('Final element of path lists must be a valid integer/frequency, got {0}'.format(path[-1]))

    def _validate_args(self, **kwargs):
        for k,v in kwargs.items():
            if v:
                try:
                    int(float(v))
                except ValueError:
                    raise BulkException('The args begin and end must be valid timestamp/numeric values - got: {0}'.format({k:v}))




class GetBulkRawData(GetBulkData):
    """Class to Get raw data to rest api."""
    _p_type = 'RawData'
    wrn = GetRawDataWarning

class GetBulkBaseRate(GetBulkData):
    """Class to Get base rate deltas to rest api."""
    _p_type = 'BaseRate'
    wrn = GetBaseRateWarning

"""Encapsulation objects for the returned data.  Subclasses the sibling
classes in the api.client.snmp module and overrides to get rid of the 
utc timestamp coersion since this is all in ms."""

class TimeSeriesDataPayload(DataPayload):
    def __init__(self, data={'data':[]}):
        super(TimeSeriesDataPayload, self).__init__(data)

    @property
    def begin_time(self):
        return self._data.get('begin_time', None)

    @property
    def end_time(self):
        return self._data.get('end_time', None)

    @property
    def data(self):
        """Return internal data from payload as list of DataPoint."""
        return [TimeSeriesDataPoint(**x) for x in self._data.get('data', [])]

class TimeSeriesDataPoint(object):
    """Class to encapsulate the returned data points."""
    __slots__ = ['ts', 'val']
    def __init__(self, ts, val):
        super(TimeSeriesDataPoint, self).__init__()
        self.ts = ts
        self.val = val

    def __repr__(self):
        return '<DataPoint: ts:{0} val:{1}>'.format(self.ts, self.val)

class TimeSeriesBulkDataPayload(object):
    def __init__(self, data={'data':[]}):
        super(TimeSeriesBulkDataPayload, self).__init__()
        self._data = data

    @property
    def begin_time(self):
        return self._data.get('begin_time', None)

    @property
    def end_time(self):
        return self._data.get('end_time', None)

    @property
    def cf(self):
        return self._data.get('cf', None)

    @property
    def data(self):
        return [TimeSeriesBulkDataRow(x) for x in self._data.get('data', [])]

    def __repr__(self):
        return '<TimeSeriesBulkDataPayload paths:{0} b:{1} e:{2}>'.format(
            len(self._data.get('data', [])), self.begin_time, self.end_time)

class TimeSeriesBulkDataRow(object):
    """docstring for TimeSeriesBulkDataRow"""
    def __init__(self, row={}):
        super(TimeSeriesBulkDataRow, self).__init__()
        self._path = row.get('path', [])
        self._data = row.get('data', [])

    @property
    def path(self):
        return self._path

    @property
    def data(self):
        return [TimeSeriesDataPoint(**x) for x in self._data]

    def __repr__(self):
        return '<TimeSeriesBulkDataRow: path:{0} len:{1}>'.format(self.path, len(self._data))
        








import calendar
import datetime
import json
import pprint
import requests
import time
import warnings

from ..util import add_apikey_header

MAX_DATETIME = datetime.datetime.max - datetime.timedelta(2)
MAX_EPOCH = calendar.timegm(MAX_DATETIME.utctimetuple())

class NodeInfoWarning(Warning): pass
class MetadataWarning(NodeInfoWarning): pass
class EventTypeWarning(NodeInfoWarning): pass
class SummaryWarning(NodeInfoWarning): pass
class DataPayloadWarning(NodeInfoWarning): pass
class DataPointWarning(NodeInfoWarning): pass
class DataHistogramWarning(NodeInfoWarning): pass
class ApiFiltersWarning(Warning): pass
class ApiConnectWarning(Warning): pass

class NodeInfo(object):
    wrn = NodeInfoWarning
    """Base class for encapsulation objects"""
    def __init__(self, data, api_url, filters):
        super(NodeInfo, self).__init__()
        self._data = data
        self.api_url = api_url
        if self.api_url: self.api_url = api_url.rstrip('/')
        self.filters = filters

        self.request_headers = {}

        self._pp = pprint.PrettyPrinter(indent=4)

    def _convert_to_datetime(self, d):
        if int(d) > MAX_EPOCH:
            return MAX_DATETIME
        else:
            return datetime.datetime.utcfromtimestamp(int(d))

    @property
    def dump(self):
        """Dump the returned/wrapped json as a pretty printed string.
        Just for debugging."""
        return self._pp.pformat(self._data)

    def http_alert(self, r):
        """
        Issue a subclass specific alert in the case that a call to the REST
        api does not return a 200 status code.
        """
        warnings.warn('Request for {0} got status: {1} - response: {2}'.format(r.url,r.status_code,r.content), self.wrn, stacklevel=2)

    def warn(self, m):
        warnings.warn(m, self.wrn, stacklevel=2)

    def inspect_request(self, r):
        if self.filters.verbose:
            print '[url: {0}]'.format(r.url)

class Metadata(NodeInfo):
    wrn = MetadataWarning
    """Class to encapsulate a metadata object.  It exposes the 
    information returned in the metadata json payload from the api 
    as a series of read-only properties.

    Will also return the associated/wrapped event-types as EventType
    objects."""
    def __init__(self, data, api_url, filters):
        super(Metadata, self).__init__(data, api_url, filters)

    @property
    def destination(self):
        return self._data.get('destination', None)

    @property
    def event_types(self):
        """Returns a list of the event-types associated with 
        the returned metadata as a list of strings."""
        e_t = []
        for e in self._data.get('event-types', []):
            e_t.append(e['event-type'])
        return e_t

    @property
    def input_destination(self):
        return self._data.get('input-destination', None)

    @property
    def input_source(self):
        return self._data.get('input-source', None)

    @property
    def ip_packet_interval(self):
        return self._data.get('ip-packet-interval', None)

    @property
    def ip_transport_protocol(self):
        return self._data.get('ip-transport-protocol', None)

    @property
    def measurement_agent(self):
        return self._data.get('measurement-agent', None)

    @property
    def metadata_count_total(self):
        return self._data.get('metadata-count-total', None)

    @property
    def metadata_key(self):
        return self._data.get('metadata-key', None)

    @property
    def sample_bucket_width(self):
        return self._data.get('sample-bucket-width', None)

    @property
    def source(self):
        return self._data.get('source', None)

    @property
    def subject_type(self):
        return self._data.get('subject-type', None)

    @property
    def time_duration(self):
        return self._data.get('time-duration', None)

    @property
    def time_interval(self):
        return self._data.get('time-interval', None)

    @property
    def time_interval_randomization(self):
        return self._data.get('time-interval-randomization', None)

    @property
    def tool_name(self):
        return self._data.get('tool-name', None)

    @property
    def uri(self):
        return self._data.get('uri', None)
    
    @property
    def get_freeform_key_value(self, key):
        return self._data.get(key, None)

    def get_all_event_types(self):
        """Generator returning all the event-types from the metadata
        json payload wrapped in EventType objects."""
        for et in self._data.get('event-types', []):
            yield EventType(et, self.api_url, self.filters)

    def get_event_type(self, event_type):
        """Returns a single named event-type (as specified by the arg
        event_type) from the metadata json payload wrapped in an  
        EventType object."""
        for et in self._data.get('event-types', []):
            if et['event-type'] == event_type:
                return EventType(et, self.api_url, self.filters)
        return None

    def __repr__(self):
        return '<Metadata/{0}: uri:{1}>'.format(self.metadata_key, self.uri)

class EventType(NodeInfo):
    wrn = EventTypeWarning
    """Class to encapsulate a single event-type from the json
    payload returned from a metadata query/Metadata object.
    Exposes the json data as a series of read-only properties.

    Is also used to fetch the data associated with the event-type
    from the API.

    Will also return related summaries as a Summary object."""
    def __init__(self, data, api_url, filters):
        super(EventType, self).__init__(data, api_url, filters)

    @property
    def base_uri(self):
        return self._data.get('base-uri', None)

    @property
    def event_type(self):
        return self._data.get('event-type', None)

    @property
    def data_type(self):
        """Returns whether this data type is a histogram. Decision
        based on event prefix which always indicates histogram type"""
        
        if self.event_type is not None and self.event_type.startswith("histogram-"):
            return "histogram"
        else:
            return "unspecified"

    @property
    def summaries(self):
        """Returns a list of strings of the names of the summaries
        associated with this event-type."""
        s_t = []
        for s in self._data.get('summaries', []):
            s_t.append((s['summary-type'],s['summary-window']))
        return s_t

    def get_all_summaries(self):
        """Generator returning all the summaries from the event-type
        json payload wrapped in Summary objects."""
        for s in self._data.get('summaries', []):
            yield Summary(s, self.api_url, self.filters, self.data_type)

    def get_summary(self, s_type, s_window):
        """Returns a single named summary (as specified by the arg
        pair s_type/s_window) from the event-type json payload wrapped 
        in a Summary object."""
        for s in self._data.get('summaries', []):
            if s['summary-type'] == s_type and \
                s['summary-window'] == str(s_window):
                return Summary(s, self.api_url, self.filters, self.data_type)
        return None

    def get_data(self):
        """Void method to pull the data associated with this event-type
        from the API.  Returns a DataPayload object to calling code."""
        r = requests.get('{0}{1}'.format(self.api_url, self.base_uri),
            params=self.filters.time_filters,
            headers=self.request_headers)

        self.inspect_request(r)

        if r.status_code == 200 and \
            r.headers['content-type'] == 'application/json':
            data = json.loads(r.text)
            return DataPayload(data, self.data_type)
        else:
            self.http_alert(r)
            return DataPayload([], self.data_type)


    def __repr__(self):
        return '<EventType/{0}: uri:{1}>'.format(self.event_type, self.base_uri)

class Summary(NodeInfo):
    wrn = SummaryWarning
    """Class to encapsulate summary information.  Exposes the summary 
    json payload in the event-type as read-only properties.

    Is also used to fetch the actual summary data from the API."""
    def __init__(self, data, api_url, filters, data_type):
        super(Summary, self).__init__(data, api_url, filters)
        self._data_type = data_type

    @property
    def data_type(self):
        return self._data_type

    @property
    def summary_type(self):
        return self._data.get('summary-type', None)

    @property
    def summary_window(self):
        return self._data.get('summary-window', None)

    @property
    def uri(self):
        return self._data.get('uri', None)

    def get_data(self):
        """Void method to pull the data associated with this event-type
        from the API.  Returns a DataPayload object to calling code."""
        r = requests.get('{0}{1}'.format(self.api_url, self.uri),
            params=self.filters.time_filters,
            headers=self.request_headers)

        self.inspect_request(r)

        if r.status_code == 200 and \
            r.headers['content-type'] == 'application/json':
            data = json.loads(r.text)
            return DataPayload(data, self.data_type)
        else:
            self.http_alert(r)
            return DataPayload([], self.data_type)

    def __repr__(self):
        return '<Summary/{0}: window:{1}>'.format(self.summary_type, self.summary_window)


class DataPayload(NodeInfo):
    wrn = DataPayloadWarning
    """Class to encapsulate returned data payload.  Holds the json
    payload of timeseries/data points internally and returns the 
    discrete data as a list of DataPoint or DataHistogram objects
    as is appropriate."""
    def __init__(self, data=[], data_type=None):
        super(DataPayload, self).__init__(data, None, None)
        self._data_type = data_type

    @property
    def data_type(self):
        return self._data_type

    @property
    def data(self):
        if self.data_type == 'histogram':
            return [DataHistogram(x) for x in self._data]
        else:
            return [DataPoint(x) for x in self._data]

    @property
    def dump(self):
        return self._pp.pformat(self._data)

    def __repr__(self):
        return '<DataPayload: len:{0} type:{1}>'.format(len(self._data), self.data_type)

class DataPoint(NodeInfo):
    __slots__ = ['ts', 'val']
    wrn = DataPointWarning
    """Class to encapsulate the data points.  Represents a single
    ts/value pair where the value is a simple numeric type."""
    def __init__(self, data={}):
        super(DataPoint, self).__init__(data, None, None)
        self.ts = self._convert_to_datetime(data.get('ts', None))
        self.val = data.get('val', None)

    @property
    def ts_epoch(self):
        return calendar.timegm(self.ts.utctimetuple())

    def __repr__(self):
        return '<DataPoint: ts:{0} val:{1}>'.format(self.ts, self.val)

class DataHistogram(NodeInfo):
    __slots__ = ['ts', 'val']
    wrn = DataHistogramWarning
    """Class to encapsulate the data histograms  Represents a single
    ts/value pair where the value is histogram written as a dict."""
    def __init__(self, data={}):
        super(DataHistogram, self).__init__(data, None, None)
        self.ts = self._convert_to_datetime(data.get('ts', None))
        self.val = data.get('val', {})

    @property
    def ts_epoch(self):
        return calendar.timegm(self.ts.utctimetuple())

    def __repr__(self):
        return '<DataHistogram: ts:{0} len:{1}>'.format(self.ts, len(self.val.keys()))

class ApiFilters(object):
    wrn = ApiFiltersWarning
    """
    Class to hold filtering/query options.

    This is instantiated and the read/write properties that correspond 
    to metadata information and time ranges can be set by the user.  
    The class will then be passed to the ApiConnect class constructor 
    and the values that have been set will be used in the metadata
    query sent to the API.

    In the case where a metadata property has a dash in it (ie: 
    input-destination), the property will be named with an underscore
    (ie: input_destination) instead.

    See the perfsonar_client.rst doc for more usage information.
    """
    def __init__(self):
        super(ApiFilters, self).__init__()
        self.verbose = False

        self._metadata_filters = {}
        self._time_filters = {}

        # Other stuff
        self.auth_username = ''
        self.auth_apikey = ''

    # Metadata level search filters

    @property
    def metadata_filters(self):
        return self._metadata_filters

    def destination():
        doc = "The destination property."
        def fget(self):
            return self._metadata_filters.get('destination', None)
        def fset(self, value):
            self._metadata_filters['destination'] = value
        def fdel(self):
            del self._metadata_filters['destination']
        return locals()
    destination = property(**destination())

    def input_destination():
        doc = "The input_destination property."
        def fget(self):
            return self._metadata_filters.get('input-destination', None)
        def fset(self, value):
            self._metadata_filters['input-destination'] = value
        def fdel(self):
            del self._metadata_filters['input-destination']
        return locals()
    input_destination = property(**input_destination())

    def input_source():
        doc = "The input_source property."
        def fget(self):
            return self._metadata_filters.get('input-source', None)
        def fset(self, value):
            self._metadata_filters['input-source'] = value
        def fdel(self):
            del self._metadata_filters['input-source']
        return locals()
    input_source = property(**input_source())

    def measurement_agent():
        doc = "The measurement_agent property."
        def fget(self):
            return self._metadata_filters.get('measurement-agent', None)
        def fset(self, value):
            self._metadata_filters['measurement-agent'] = value
        def fdel(self):
            del self._metadata_filters['measurement-agent']
        return locals()
    measurement_agent = property(**measurement_agent())

    def source():
        doc = "The source property."
        def fget(self):
            return self._metadata_filters.get('source', None)
        def fset(self, value):
            self._metadata_filters['source'] = value
        def fdel(self):
            del self._metadata_filters['source']
        return locals()
    source = property(**source())

    def tool_name():
        doc = "The tool_name property."
        def fget(self):
            return self._metadata_filters.get('tool-name', None)
        def fset(self, value):
            self._metadata_filters['tool-name'] = value
        def fdel(self):
            del self._metadata_filters['tool-name']
        return locals()
    tool_name = property(**tool_name())

    # Additional metadata search criteria
    # event-type, summary-type, summary-window and subject-type

    def event_type():
        doc = "The event_type property."
        def fget(self):
            return self._metadata_filters.get('event-type', None)
        def fset(self, value):
            self._metadata_filters['event-type'] = value
        def fdel(self):
            del self._metadata_filters['event-type']
        return locals()
    event_type = property(**event_type())

    def subject_type():
        doc = "The subject_type property."
        def fget(self):
            return self._metadata_filters.get('subject-type', None)
        def fset(self, value):
            self._metadata_filters['subject-type'] = value
        def fdel(self):
            del self._metadata_filters['subject-type']
        return locals()
    subject_type = property(**subject_type())

    def summary_type():
        doc = "The summary_type property."
        def fget(self):
            return self._metadata_filters.get('summary-type', None)
        def fset(self, value):
            self._metadata_filters['summary-type'] = value
        def fdel(self):
            del self._metadata_filters['summary-type']
        return locals()
    summary_type = property(**summary_type())

    def summary_window():
        doc = "The summary_window property."
        def fget(self):
            return self._metadata_filters.get('summary-window', None)
        def fset(self, value):
            self._metadata_filters['summary-window'] = value
        def fdel(self):
            del self._metadata_filters['summary-window']
        return locals()
    summary_window = property(**summary_window())

    # Time range search filters

    def _check_time(self, ts):
        try:
            t_s = int(float(ts))
            return t_s
        except ValueError:
            self.warn('The timestamp value {0} is not a valid integer'.format(ts))

    @property
    def time_filters(self):
        return self._time_filters

    def time():
        doc = "The time property."
        def fget(self):
            return self._time_filters.get('time', None)
        def fset(self, value):
            self._time_filters['time'] = self._check_time(value)
        def fdel(self):
            del self._time_filters['time']
        return locals()
    time = property(**time())

    def time_start():
        doc = "The time_start property."
        def fget(self):
            return self._time_filters.get('time-start', None)
        def fset(self, value):
            self._time_filters['time-start'] = self._check_time(value)
        def fdel(self):
            del self._time_filters['time-start']
        return locals()
    time_start = property(**time_start())

    def time_end():
        doc = "The time_end property."
        def fget(self):
            return self._time_filters.get('time-end', None)
        def fset(self, value):
            self._time_filters['time-end'] = self._check_time(value)
        def fdel(self):
            del self._time_filters['time-end']
        return locals()
    time_end = property(**time_end())

    def time_range():
        doc = "The time_range property."
        def fget(self):
            return self._time_filters.get('time-range', None)
        def fset(self, value):
            self._time_filters['time-range'] = self._check_time(value)
        def fdel(self):
            del self._time_filters['time-range']
        return locals()
    time_range = property(**time_range())

    def warn(self, m):
        warnings.warn(m, self.wrn, stacklevel=2)

class ApiConnect(object):
    wrn = ApiConnectWarning
    """
    Core class to pull data from the rest api.

    This is the "top level" class a client program will use to 
    query the perfsonar API.  The args api_url, username and api_key 
    all have their typical usage.  The ApiFilters object will have
    the relevant query criteria assigned to it, and the get_metadata
    method will return the metadata that meet the search criteria 
    as Metadata objects.

    See the perfsonar_client.rst doc for more usage information.
    """
    def __init__(self, api_url, filters=ApiFilters(), username='', api_key='',
            script_alias='esmond'):
        super(ApiConnect, self).__init__()
        self.api_url = api_url.rstrip("/")
        self.filters = filters
        self.filters.auth_username = username
        self.filters.auth_apikey = api_key
        self.script_alias = script_alias

        if self.script_alias: 
            self.script_alias = script_alias.rstrip('/')
            self.script_alias = script_alias.lstrip('/')

        self.request_headers = {}

    def get_metadata(self):
        if self.script_alias:
            archive_url = '{0}/{1}/perfsonar/archive/'.format(self.api_url, self.script_alias)
        else:
            archive_url = '{0}/perfsonar/archive/'.format(self.api_url)

        r = requests.get(archive_url,
            params=dict(self.filters.metadata_filters, **self.filters.time_filters),
            headers = self.request_headers)

        self.inspect_request(r)

        data = list()

        if r.status_code == 200 and \
            r.headers['content-type'] == 'application/json':
            data = json.loads(r.text)
            
            if data:
                m_total = Metadata(data[0], self.api_url, self.filters).metadata_count_total
            else:
                m_total = 0

            # Check to see if we are geting paginated metadata, tastypie 
            # has a limit to how many results it will return even if 
            # ?limit=0
            if len(data) < m_total:
                # looks like we got paginated content.
                if self.filters.verbose: print 'pagination - metadata_count_total: {0} got: {1}\n'.format(m_total, len(data))
                initial_offset = len(data) # should be the tastypie internal limit of 1000
                offset = initial_offset

                while offset < m_total:
                    if self.filters.verbose:
                        print 'current total results: {0}'.format(len(data))
                        print 'issuing request with offset: {0}'.format(offset)

                    r = requests.get(archive_url,
                        params=dict(self.filters.metadata_filters, offset=offset, **self.filters.time_filters),
                        headers = self.request_headers)
                    self.inspect_request(r)

                    if r.status_code != 200:
                        print 'error fetching paginated content'
                        self.http_alert(r)
                        return
                        yield

                    tmp = json.loads(r.text)

                    if self.filters.verbose: print 'got {0} results\n'.format(len(tmp))

                    data.extend(tmp)
                    offset += initial_offset

            if self.filters.verbose: print 'final result count: {0}\n'.format(len(data))

            for i in data:
                yield Metadata(i, self.api_url, self.filters)
        else:
            self.http_alert(r)
            return
            yield

    def inspect_request(self, r):
        if self.filters.verbose:
            print '[url: {0}]'.format(r.url)

    def inspect_payload(self, p):
        if self.filters.verbose > 1:
            print '[POST payload: {0}]'.format(json.dumps(p, indent=4))

    def http_alert(self, r):
        warnings.warn('Request for {0} got status: {1} - response: {2}'.format(r.url,r.status_code, r.content), self.wrn, stacklevel=2)

    def warn(self, m):
        warnings.warn(m, self.wrn, stacklevel=2)





"""
Utility code for perfsonar esmond client programs.
"""

import calendar
import copy
import cStringIO
import csv
import datetime
import json
import os
import socket
import sys
import urllib
import urlparse

from optparse import OptionParser
from collections import OrderedDict
from dateutil.parser import parse

from .query import ApiFilters, DataPayload

# Event types with an associated "formatting type" to be 
# used by esmond-get, etc.
EVENT_MAP = OrderedDict([
    ('failures', 'failures'),
    ('histogram-owdelay', 'histogram'),
    ('histogram-rtt', 'histogram'),
    ('histogram-ttl', 'histogram'),
    ('histogram-ttl-reverse', 'histogram'),
    ('ntp-delay', 'numeric'),
    ('ntp-dispersion', 'numeric'),
    ('ntp-jitter', 'numeric'),
    ('ntp-offset', 'numeric'),
    ('ntp-polling-interval', 'numeric'),
    ('ntp-reach', 'numeric'),
    ('ntp-stratum', 'numeric'),
    ('ntp-wander', 'numeric'),
    ('packet-duplicates', 'numeric'),
    ('packet-duplicates-bidir', 'numeric'),
    ('packet-loss-rate', 'numeric'),
    ('packet-loss-rate-bidir', 'numeric'),
    ('packet-trace', 'packet_trace'),
    ('packet-count-lost', 'numeric'),
    ('packet-count-lost-bidir', 'numeric'),
    ('packet-count-sent', 'numeric'),
    ('packet-reorders', 'numeric'),
    ('packet-reorders-bidir', 'numeric'),
    ('packet-retransmits', 'numeric'),
    ('packet-retransmits-subintervals', 'subintervals'),
    ('path-mtu', 'numeric'),
    ('streams-packet-retransmits', 'number_list'),
    ('streams-packet-retransmits-subintervals', 'subinterval_list'),
    ('streams-throughput', 'number_list'),
    ('streams-throughput-subintervals', 'subinterval_list'),
    ('throughput', 'numeric'),
    ('throughput-subintervals', 'subintervals'),
    ('time-error-estimates', 'numeric'),
])

EVENT_TYPES = EVENT_MAP.keys()

def event_format(et):
    return EVENT_MAP[et]

DEFAULT_FIELDS = [
        'source', 
        'destination', 
        'measurement_agent',
        'input_source',
        'input_destination',
        'tool_name', 
]

class HeaderRow(OrderedDict):
    pass

#
# Exceptions for client operations
#

class EsmondClientException(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)

class EsmondClientWarning(Warning): pass

#
# Output classes for clients
#

class EsmondOutput(object):
    """Base class for output classes."""

    def __init__(self, data, columns):
        self._data = data
        self._columns = columns
        self._output = None

        self._list_fields = None

        if not isinstance(self._data, list):
            raise EsmondClientException('Data arg must be a list')

        if len(self._data) and not isinstance(self._data[0], dict):
            raise EsmondClientException('Data arg must be a list of dicts')

    def get_output(self):
        raise NotImplementedError('Implement in subclasses.')

    def has_data(self):
        if len(self._data):
            return True
        else:
            return False

    def add_to_payload(self, d):
        if not isinstance(d, list):
            raise EsmondClientException('Arg to add_to_payload must be a list')

        self._data.extend(d)

    def _massage_row_dict(self, d):
        # scan first instance to see if we need to fix anything - 
        # no point in processing each row if not necessary.
        if self._list_fields == None:
            self._list_fields = []
            for k,v in d.items():
                if isinstance(v, list):
                    self._list_fields.append(k)

        # if no changes need to be made, just quit
        if len(self._list_fields) == 0:
            return d

        # don't change the original data
        new_d = copy.copy(d)

        # turn any lists into comma separated sequences
        for lf in self._list_fields:
            new_d[lf] = ', '.join( [ str(x) for x in new_d.get(lf) ] )

        return new_d


class HumanOutput(EsmondOutput):
    """Output for human/console output."""

    def __init__(self, data, columns, extended_data=False):
        super(HumanOutput, self).__init__(data, columns)

        self._extended_data = extended_data

    def get_output(self):
        entry_delim = '= + = + = + = + = + =\n'

        if not self._output:
            self._output = ''
            if self.has_data():
                for row in self._data:
                    if isinstance(row, HeaderRow):
                        for k,v in row.items():
                            self._output += '{0}: {1}\n'.format(k, v)
                    else:
                        row = self._massage_row_dict(row)
                        for c in self._columns:
                            self._output += '{0}: {1}\n'.format(c, row.get(c))
                        if self._extended_data:
                            for k,v in row.items():
                                if k in self._columns: continue
                                self._output += '{0}: {1}\n'.format(k,v)
                    self._output += entry_delim
                self._output = self._output[:self._output.rfind(entry_delim)]
            else:
                self._output = 'No data found.'

        return self._output

class JsonOutput(EsmondOutput):
    """Output results as a json blob."""

    def get_output(self):
        if not self._output:
            if self.has_data():
                self._output = json.dumps(self._data)
            else:
                self._output = json.dumps( [ {'msg': 'No data found.'} ] )
        return self._output

class CSVOutput(EsmondOutput):
    """Format output for CSV."""
    def get_output(self):
        if not self._output:
            cfile = cStringIO.StringIO()

            writer = csv.DictWriter(cfile, fieldnames=self._columns, extrasaction='ignore')
            writer.writeheader()
            if self.has_data():
                for row in self._data:
                    writer.writerow(self._massage_row_dict(row))
            else:
                d = dict( [ (x, 'No data') for x in self._columns ] )
                writer.writerow(d)

            self._output = cfile.getvalue()
            cfile.close()
        return self._output

def output_factory(options, data, columns):
    """
    Factory function to hand the appropriate output class 
    back to the client programs.
    """
    if options.format == 'human':
        if not options.metadata:
            return HumanOutput(data, columns)
        else:
            return HumanOutput(data, columns, extended_data=True)
    elif options.format == 'json':
        return JsonOutput(data, None)
    elif options.format == 'csv':
        return CSVOutput(data, columns)

class HostnameConversion(object):
    def __init__(self, options):
        self._options = options
        self._ns_cache = {}

        self._ip_fields = ['source', 'destination', 'ip', 'measurement_agent']

    def convert(self, d):
        if not self._options.ip:
            for i in self._ip_fields:
                if d.get(i):
                    ip = d.get(i)
                    if not self._ns_cache.get(ip):
                        self._ns_cache[ip] = socket.getfqdn(ip)
                    d[i] = self._ns_cache.get(ip)
        return d

#
# Generate output for different event types.
# 

def data_format_factory(options, seed_bulk_output=False):
    """
    Factory function to format the actual data for output.

    The appropriate sub-function is identified by the format_map.
    All functions hand back a header list of the appropriate columns
    for that measurment type and a list of dicts with corresponding keys.
    Both of which are passed off to the output rendering code.
    """

    # these columns/values are common to all measurements.
    header_base = [
        'source', 
        'destination', 
        'event_type', 
        'tool', 
        'summary_type', 
        'summary_window',
        'timestamp'
    ]

    ip_convert = HostnameConversion(options)

    def get_summary_type():
        if not options.summary_type:
            return 'base'
        else:
            return options.summary_type

    def get_payload(et):
        if not options.summary_type:
            # unsummarized data
            return et.get_data()
        else:
            # summary data
            s = et.get_summary(options.summary_type, options.summary_window)
            if not s: 
                return DataPayload()
            else:
                return s.get_data()

    def massage_output(d):
        """any modifications to the data dicts here."""

        # ip -> hostname if need be
        d = ip_convert.convert(d)

        return d

    def header_row(m, dp):
        """Special row that will be handled by the human readable 
        output class."""
        header = [
            ('source', m.source),
            ('destination', m.destination),
            ('event_type', options.type),
            ('tool', m.tool_name),
            ('summary_type', get_summary_type()),
            ('summary_window', options.summary_window),
            ('timestamp', str(dp.ts)),
        ]
        return HeaderRow(header)
        

    def format_numeric(conn):
        """aggregation, 300, 3600, 86400"""

        header = header_base + ['value']

        data = list()

        if seed_bulk_output: return header, data

        for m in conn.get_metadata():
            et = m.get_event_type(options.type)

            for dp in get_payload(et).data:
                d = dict(
                        source=m.source,
                        destination=m.destination,
                        event_type=options.type,
                        tool=m.tool_name,
                        summary_type=get_summary_type(),
                        summary_window=options.summary_window,
                        timestamp=str(dp.ts),
                        value=dp.val,
                    )
                data.append(massage_output(d))

        return header, data

    def format_failures(conn):

        header = header_base + ['msg']

        data = list()

        if seed_bulk_output: return header, data

        for m in conn.get_metadata():
            et = m.get_event_type(options.type)
            for dp in et.get_data().data:
                d = dict(
                    source=m.source,
                    destination=m.destination,
                    event_type=options.type,
                    tool=m.tool_name,
                    summary_type=get_summary_type(),
                    summary_window=options.summary_window,
                    timestamp=str(dp.ts),
                    msg=dp.val.get('error')
                )
                data.append(massage_output(d))

        return header, data

    def format_packet_trace(conn):

        test_header = ['ttl', 'query', 'success', 'ip', 'rtt', 'mtu', 'error_message']

        if options.format != 'human':
            header = header_base + test_header
        else:
            header = test_header

        data = list()

        if seed_bulk_output: return header, data

        for m in conn.get_metadata():
            et = m.get_event_type(options.type)
            for dp in et.get_data().data:
                if options.format == 'human':
                    data.append(massage_output(header_row(m,dp)))
                for val in dp.val:
                    if options.format != 'human':
                        d = dict(
                            source=m.source,
                            destination=m.destination,
                            event_type=options.type,
                            tool=m.tool_name,
                            summary_type=get_summary_type(),
                            summary_window=options.summary_window,
                            timestamp=str(dp.ts),
                            ttl=val.get('ttl'),
                            query=val.get('query'),
                            success=val.get('success'),
                            ip=val.get('ip'),
                            rtt=val.get('rtt'),
                            mtu=val.get('mtu'),
                            error_message=val.get('error_message')
                        )
                    else:
                        d = dict(
                            ttl=val.get('ttl'),
                            query=val.get('query'),
                            success=val.get('success'),
                            ip=val.get('ip'),
                            rtt=val.get('rtt'),
                            mtu=val.get('mtu'),
                            error_message=val.get('error_message')
                        )
                    data.append(massage_output(d))

        return header, data

    def format_histogram(conn):
        """aggregation, statistics, 300, 3600, 86400"""

        if options.summary_type == 'statistics':
            header = header_base + ['min', 'median', 'max', 
                'mean', 'mode', 'standard_deviation', 'variance', 
                'percentile_25', 'percentile_75', 'percentile_95']
        else:
            header = header_base + ['bucket', 'value']

        data = list()

        if seed_bulk_output: return header, data

        for m in conn.get_metadata():
            et = m.get_event_type(options.type)

            for dp in get_payload(et).data:
                if options.summary_type == 'statistics':
                    d = dict(
                        source=m.source,
                        destination=m.destination,
                        event_type=options.type,
                        tool=m.tool_name,
                        summary_type=get_summary_type(),
                        summary_window=options.summary_window,
                        timestamp=str(dp.ts),
                        min=dp.val.get('minimum'),
                        median=dp.val.get('median'),
                        max=dp.val.get('maximum'),
                        mean=dp.val.get('mean'),
                        mode=dp.val.get('mode'),
                        standard_deviation=dp.val.get('standard-deviation'),
                        variance=dp.val.get('variance'),
                        percentile_25=dp.val.get('percentile-25'),
                        percentile_75=dp.val.get('percentile-75'),
                        percentile_95=dp.val.get('percentile-95'),

                    )
                else:
                    d = dict(
                        source=m.source,
                        destination=m.destination,
                        event_type=options.type,
                        tool=m.tool_name,
                        summary_type=get_summary_type(),
                        summary_window=options.summary_window,
                        timestamp=str(dp.ts),
                        bucket=m.sample_bucket_width,
                        value=dp.val
                    )
                data.append(massage_output(d))

        return header, data

    def format_subintervals(conn):
        
        header = header_base + ['start', 'duration', 'value']

        data = list()

        if seed_bulk_output: return header, data

        for m in conn.get_metadata():
            et = m.get_event_type(options.type)

            for dp in et.get_data().data:
                for val in dp.val:
                    d = dict(
                        source=m.source,
                        destination=m.destination,
                        event_type=options.type,
                        tool=m.tool_name,
                        summary_type=get_summary_type(),
                        summary_window=options.summary_window,
                        timestamp=str(dp.ts),
                        start=val.get('start'),
                        duration=val.get('duration'),
                        value=val.get('val'),
                    )
                    data.append(massage_output(d))

        return header, data

    def format_number_list(conn):

        header = header_base + ['stream_num', 'value']

        data = list()

        if seed_bulk_output: return header, data

        for m in conn.get_metadata():
            et = m.get_event_type(options.type)
            for dp in et.get_data().data:
                for i in range(len(dp.val)):
                    d = dict(
                        source=m.source,
                        destination=m.destination,
                        event_type=options.type,
                        tool=m.tool_name,
                        summary_type=get_summary_type(),
                        summary_window=options.summary_window,
                        timestamp=str(dp.ts),
                        stream_num=i,
                        value=dp.val[i],
                    )
                    data.append(massage_output(d))

        return header, data

    def format_subinterval_list(conn):

        header = header_base + ['stream_num', 'start', 'duration', 'value']

        data = list()

        if seed_bulk_output: return header, data

        for m in conn.get_metadata():
            et = m.get_event_type(options.type)
            for dp in et.get_data().data:
                for stream_num in range(len(dp.val)):
                    for i in range(len(dp.val[stream_num])):
                        d = dict(
                            source=m.source,
                            destination=m.destination,
                            event_type=options.type,
                            tool=m.tool_name,
                            summary_type=get_summary_type(),
                            summary_window=options.summary_window,
                            timestamp=str(dp.ts),
                            stream_num=stream_num,
                            start=dp.val[stream_num][i].get('start'),
                            duration=dp.val[stream_num][i].get('duration'),
                            value=dp.val[stream_num][i].get('val'),
                        )
                        data.append(massage_output(d))

        return header, data

    format_map = dict(
        failures=format_failures,
        histogram=format_histogram,
        number_list=format_number_list,
        numeric=format_numeric,
        packet_trace=format_packet_trace,
        subintervals=format_subintervals,
        subinterval_list=format_subinterval_list,
    )

    return format_map.get(event_format(options.type))

# 
# Generate a filename and file handle for output.
#

def get_outfile(options, metadata, event_type):

    if not options.ip:
        source = socket.getfqdn(metadata.source)
        dest = socket.getfqdn(metadata.destination)
    else:
        source = metadata.source
        dest = metadata.destination

    def utciso(dt):
        return datetime.datetime.utcfromtimestamp(dt).strftime('%Y-%m-%d')

    s = utciso(metadata.filters.time_start)
    e = utciso(metadata.filters.time_end)

    outfile =  '{0}.{1}'.format('_'.join([source, dest, event_type, s, e]), options.format)

    return open(os.path.join(os.path.abspath(options.output_dir), outfile), 'wb')

#
# Command line argument validation functions
#

def check_url(options, parser):
    if not options.url:
        print '--url is a required arg\n'
        parser.print_help()
        sys.exit(-1)

    # trim URI from --url since people will cut and paste from 
    # the list of MAs.
    url = options.url
    up = urlparse.urlparse(url)
    if up.path.startswith('/esmond/perfsonar/archive'):
        options.url = url.replace(up.path, '')
        print '\n not necessary to add /esmond/perfsonar/archive to --url arg - trimming'

    try:
        urllib.urlopen(options.url)
    except Exception, e:
        print 'Could not open --url {0} - error: {1}'.format(options.url, e)

def check_valid_hostnames(options, parser, hn_args=[]):
    try:
        for hn in hn_args:
            if getattr(options, hn):
                socket.gethostbyname(getattr(options, hn))
    except:
        print '--{0} arg had invalid hostname: {1}'.format(hn, getattr(options, hn))
        sys.exit(-1)

def check_event_types(options, parser, require_event):
    if options.type and options.type not in EVENT_TYPES:
        print '{0} is not a valid event type'.format(options.type)
        list_event_types()
        sys.exit(-1)
    if require_event and not options.type:
        print 'The --event-type arg is required. Use -L to see a list.\n'
        parser.print_help()
        sys.exit(-1)

def check_formats(options, parser):
    f_args = ['human', 'json', 'csv']
    if options.format not in f_args:
        print '{0} is not a valid --output-format arg (one of: {1})'.format(options.format, f_args)
        sys.exit(-1)
    if options.format == 'csv' and options.metadata:
        print '--output-format csv can not be used with --metadata-extended'
        sys.exit(-1)

def check_summary(options, parser):
    s_args = ['aggregation', 'average', 'statistics']
    if options.summary_type and options.summary_type not in s_args:
        print '{0} is not a valid --summary-type arg (one of: {1})'.format(options.summary_type, s_args)
        sys.exit(-1)

def src_dest_required(options, parser):
    if not options.src or not (options.dest or (options.type is not None and options.type.startswith('ntp-'))):
        print '--src and --dest args are both required (ntp-* event types only require --src).\n'
        parser.print_help()
        sys.exit(-1)

def valid_output_dir(options, parser):
    if options.format == 'human':
        print 'please specify either json or csv --output-format for bulk output.\n'
        parser.print_help()
        sys.exit(-1)
    if not os.path.exists(os.path.abspath(options.output_dir)):
        print 'output path {0} does not exist.'.format(os.path.abspath(options.output_dir))
        sys.exit(-1)

#
# Misc command line --arg functions.
#

def list_event_types():
    print '\nValid event types:'
    for et in EVENT_TYPES:
        print '    {0}'.format(et)

#
# Utility functions to import into clients.
#

def get_start_and_end_times(options):
    """
    See:
    https://dateutil.readthedocs.org/en/latest/examples.html#parse-examples
    To see the variety of date formats that it will accept.
    """
    start = end = None

    if not options.start:
        start = datetime.datetime.utcnow() - datetime.timedelta(hours=24)
    else:
        try:
            start = parse(options.start)
        except:
            print 'could not parse --start-time arg: {0}'.format(options.start)
            sys.exit(-1)

    if not options.end:
        end = datetime.datetime.utcnow().replace(microsecond=0)
    else:
        try:
            end = parse(options.end)
        except:
            print 'could not parse --end-time arg: {0}'.format(options.end)
            sys.exit(-1)

    return start, end

#
# Canned option parsers for clients
#

def perfsonar_client_opts(require_src_dest=False, require_event=False,
    require_output=False):
    """
    Return a standard option parser for the perfsonar clients.
    """
    usage = '%prog [ -u URL -s SRC -d DEST | -a AGENT | -e EVENT | -t TOOL | -L | -o FORMAT | -v ]'
    usage += '\n--begin and --end args parsed by python-dateutil so fairly flexible with the date formats.'
    parser = OptionParser(usage=usage)
    parser.add_option('-u', '--url', metavar='URL',
            type='string', dest='url', 
            help='URL of esmond API you want to talk to.')
    parser.add_option('-s', '--src', metavar='SRC',
            type='string', dest='src', 
            help='Host originating the test.')
    parser.add_option('-d', '--dest', metavar='DEST',
            type='string', dest='dest', 
            help='Test endpoint.')
    parser.add_option('-a', '--agent', metavar='AGENT',
            type='string', dest='agent', 
            help='Host that initiated the test - useful for central MAs.')
    parser.add_option('-e', '--event-type', metavar='EVENT',
            type='string', dest='type', 
            help='Type of data (loss, latency, throughput, etc) - see -L arg.')
    parser.add_option('-t', '--tool', metavar='TOOL',
            type='string', dest='tool', 
            help='Tool used to run test (bwctl/iperf3, powstream, "bwctl/tracepath,traceroute", gridftp, etc).')
    parser.add_option('-S', '--start-time', metavar='START',
            type='string', dest='start', 
            help='Start time of query (default: 24 hours ago).')
    parser.add_option('-E', '--end-time', metavar='END',
            type='string', dest='end', 
            help='End time of query (default: now).')
    parser.add_option('-F', '--filter', metavar='FILTER',
            type='string', dest='filter', action='append',
            help='Specify additional query filters - format: -F key:value. Can be used multiple times, invalid filters will be ignored.')
    parser.add_option('-L', '--list-events',
            dest='list_event', action='store_true', default=False,
            help='List available event types.')
    parser.add_option('-M', '--metadata-extended',
            dest='metadata', action='store_true', default=False,
            help='Show extended metadata tool-specific values (can not be used with -o csv).')
    parser.add_option('-T', '--summary-type', metavar='SUMMARY_TYPE',
            type='string', dest='summary_type', 
            help='Request summary data of type [aggregation, average, statistics].')
    parser.add_option('-W', '--summary-window', metavar='SUMMARY_WINDOW',
            type='int', dest='summary_window', default=0,
            help='Timeframe in seconds described by the summary (default: %default).')
    parser.add_option('-o', '--output-format', metavar='O_FORMAT',
            type='string', dest='format', default='human',
            help='Output format [human, json, csv] (default: human).')
    if require_output:
        parser.add_option('-D', '--output-directory', metavar='DIR',
                type='string', dest='output_dir', default=os.getcwd(),
                help='Directory to output files to (default: %default).')
    parser.add_option('-I', '--ip',
            dest='ip', action='store_true', default=False,
            help='Show source/dest as IP addresses, not hostnames.')
    parser.add_option('-v', '--verbose',
        dest='verbose', action='store_true', default=False,
        help='Verbose output.')
    options, args = parser.parse_args()

    if options.list_event:
        list_event_types()
        sys.exit(0)

    check_url(options, parser)

    if require_src_dest:
        src_dest_required(options, parser)

    if require_output:
        valid_output_dir(options, parser)

    check_valid_hostnames(options, parser, hn_args=['src', 'dest', 'agent'])

    check_event_types(options, parser, require_event)

    check_summary(options, parser)

    check_formats(options, parser)

    return options, args

def perfsonar_client_filters(options):
    """
    Return a standard filter object based on the opts in 
    perfsonar_client_opts()
    """

    start, end = get_start_and_end_times(options)

    filters = ApiFilters()
    filters.source = options.src
    filters.destination = options.dest
    filters.measurement_agent = options.agent
    filters.event_type = options.type
    filters.time_start = calendar.timegm(start.utctimetuple())
    filters.time_end = calendar.timegm(end.utctimetuple())
    filters.tool_name = options.tool
    filters.summary_type = options.summary_type
    if options.summary_window:
        filters.summary_window = options.summary_window
    filters.verbose = options.verbose

    if options.filter:
        # Apply arbritrary metadata filters
        for f in options.filter:
            if f.find(':') == -1:
                print '--filter arg {0} should be of the format key:value'.format(f)
                continue
            k,v = f.split(':')
            key = k.replace('-', '_')
            if not hasattr(filters, k):
                print '--filter arg {0} is not a valid filtering value'.format(key)
                continue
            setattr(filters, key, v)

    return filters




import calendar
import datetime
import json
import pprint
import requests
import time
import warnings

from .util import add_apikey_header, AlertMixin

"""
Library to fetch data from 'simplified' API /v1/snmp/ namespace.

The class ApiConnect is the 'entry point' that the client uses, and 
the ApiFilters class is used to set time/device/etc filters, and is 
passed to the ApiConnect class as an argument.

Example:

    filters = ApiFilters()

    if options.last:
        filters.begin_time = int(time.time() - (options.last*60))

    if options.devices:
        filters.devices = options.devices

    conn = ApiConnect(options.api_url, filters)

After the entry point is set up, it returns device objects, which in 
turn return associated interface objects, endpoint objects and data 
point objects.  

Traversing down through hierarchy of objects:

    for d in conn.get_devices():
        for i in d.get_interfaces():
            for e in i.get_endpoints():
                payload_object = e.get_data()
                    list_of_data_point_objects = payload_object.data
                        for i in list_of_data_point_objects:
                            print i.ts, i.val

Where timestamp values are returned, they are python datetime objects.

NOTE: actually executing that example w/out filtering/limiting the 
devices being polled will grab a lot of data from the API so proceed 
with caution!

The objects that encapsulate the device/interface/etc data that are
returned from the api follow the general pattern where information in 
a given object are accessed by properties, and when acutal 'work' is 
being done (hitting the api), that is done by a method prefixed with
get_ (ex: get_interfaces()).
"""

MAX_DATETIME = datetime.datetime.max - datetime.timedelta(2)
MAX_EPOCH = calendar.timegm(MAX_DATETIME.utctimetuple())

class NodeInfoWarning(Warning): pass
class DeviceWarning(NodeInfoWarning): pass
class InterfaceWarning(NodeInfoWarning): pass
class EndpointWarning(NodeInfoWarning): pass
class DataPayloadWarning(NodeInfoWarning): pass
class BulkDataPayloadWarning(NodeInfoWarning): pass
class ApiFiltersWarning(Warning): pass
class ApiConnectWarning(Warning): pass
class ApiNotFound(Exception): pass
class DeviceNotFound(ApiNotFound): pass
class DeviceCollectionNotFound(ApiNotFound): pass
class InterfaceNotFound(ApiNotFound): pass
class EndpointNotFound(ApiNotFound): pass

# - Encapsulation classes for nodes (device, interface, etc).

class NodeInfo(AlertMixin, object):
    wrn = NodeInfoWarning
    """Base class for encapsulation objects"""
    def __init__(self, data, api_url, filters):
        super(NodeInfo, self).__init__()
        self._data = data
        self.api_url = api_url
        if self.api_url:
            self.api_url = api_url.rstrip("/")
        self.filters = filters

        self.request_headers = {}

        # Default rest parameters for all objects
        if self.filters:
            self._default_filters = self.filters.default_filters
            if self.filters.auth_username and self.filters.auth_apikey:
                add_apikey_header(self.filters.auth_username, 
                    self.filters.auth_apikey, self.request_headers)

        self.pp = pprint.PrettyPrinter(indent=4)

    def _convert_to_datetime(self, d):
        """API returns both unix timestamps and ISO time so
        normalize to datetime objects transparently"""
        t = None
        
        try:
            i = int(d)
            if i > MAX_EPOCH:
                # bullet proof against out of range datetime errors
                t = max_datetime
            else:
                t = datetime.datetime.utcfromtimestamp(i)
        except ValueError:
            # Not an epoch timestamp
            pass

        if not t:
            # Presume this is ISO time
            t = datetime.datetime.strptime(d, "%Y-%m-%dT%H:%M:%S.%f")

        return t

    # Properties for most subclasses
    @property
    def begin_time(self):
        if self._data.get('begin_time', None):
            return self._convert_to_datetime(self._data.get('begin_time'))
        else:
            return None

    @property
    def end_time(self):
        if self._data.get('end_time', None):
            return self._convert_to_datetime(self._data.get('end_time'))
        else:
            return None

    @property
    def id(self):
        return self._data.get('id', None)

    @property
    def leaf(self):
        return self._data.get('leaf', None)

    @property
    def resource_uri(self):
        return self._data.get('resource_uri', None)

    @property
    def children(self):
        children = []
        for child in self._data.get('children'):
            children.append(dict(name=child.get('name'), label=child.get('name'), leaf=child.get('leaf')))

        return children

    # Utility properties
    @property
    def dump(self):
        return self.pp.pformat(self._data)

    def inspect_request(self, r):
        if self.filters.verbose:
            print '[url: {0}]'.format(r.url)


class Device(NodeInfo):
    wrn = DeviceWarning
    """Class to encapsulate device information"""
    def __init__(self, data, api_url, filters):
        super(Device, self).__init__(data, api_url, filters)

    # Property attrs unique to devices
    @property
    def active(self):
        return self._data.get('active', None)

    @property
    def name(self):
        return self._data.get('name', None)

    @property
    def oidsets(self):
        # Return a copy of the internal list so we don't have unintended
        # changes to the internal payload by reference.
        return self._data.get('oidsets', None)[:]

    def get_interfaces(self, **filters):
        """
        Issue a call to the API to get a list of interfaces (via a 
        generator) associated with the device this object refers to.
        """
        uri = None
        for c in self._data['children']:
            if c['name'] == 'interface':
                uri = c['uri']
                break

        if uri:
            r = requests.get('{0}{1}/'.format(self.api_url, uri),
                params=self.filters.compose_filters(filters),
                headers=self.request_headers)

            self.inspect_request(r)

            if r.status_code == 200 and \
                r.headers['content-type'] == 'application/json':
                data = json.loads(r.text)
                for i in data['children']:
                    yield Interface(i, self.api_url, self.filters)
            else:
                self.http_alert(r)
                return
                yield

    def get_interface(self, ifName):
        ifaces = list(self.get_interfaces(ifName=ifName))
        if len(ifaces) == 0:
            if self.filters.verbose > 1:
                print "[bad interface: {0}]".format(ifName)
            raise InterfaceNotFound()

        return ifaces[0]

    def get_device_collection(self, name):
        for child in self._data.get('children'):
            if child['name'] == name:
                return DeviceCollection(self, name, self.filters)

        raise DeviceCollectionNotFound(name)

    @property
    def children(self):
        return [dict(name='interface', label='interface', leaf=False)]

    def get_child(self, args, path=[]):
        name = args[0]
        args = args[1:]

        if self.filters.verbose > 1:
            print "[Device.get_child {0} {1} {2}]".format(name, args, path)

        try:
            collection = self.get_device_collection(name)
        except DeviceCollectionNotFound:
            return None

        path.append(name)
        if args:
            return collection.get_child(args, path=path)
        else:
            return collection

    def set_oidsets(self, oidsets):
        if not oidsets or not isinstance(oidsets, list):
            self.warn('oidsets arg must be a non-empty list')
            return

        if not self.filters.auth_username or not self.filters.auth_apikey:
            self.warn('Must supply username and api key args to alter oidsets for a device - aborting set_oidsets().')
            return

        r = requests.get('{0}/v1/oidset/'.format(self.api_url),
            headers=self.request_headers)
        if r.status_code != 200:
            self.warn('Could not get a list of valid oidsets from {0}/v1/oidset/ - aborting'.format(self.api_url) )
            return
        valid_oidsets = json.loads(r.content)

        for o in oidsets:
            if o not in valid_oidsets:
                self.warn('{0} is not a valid oidset - aborting oidset update.')

        self._data['oidsets'] = oidsets

        self.request_headers['content-type'] = 'application/json'

        p = requests.put('{0}{1}'.format(self.api_url, self.resource_uri), 
            data=json.dumps(self._data), headers=self.request_headers)

        if p.status_code != 204:
            self.warn('Did not successfully change oidsets - might be an auth problem.')
            self.http_alert(p)



    def __repr__(self):
        return '<Device/{0}: uri:{1}>'.format(self.name, self.resource_uri)

class DeviceCollection(object):
    def __init__(self, device, name, filters):
        self.device = device
        self.name = name
        self.filters = filters

    @property
    def leaf(self):
        return False

    @property
    def children(self):
        if self.name == "interface":
            return [ dict(path="", name=i.ifName, label="%s %s" % (i.ifName, i.ifAlias)) for i in self.device.get_interfaces() ]
        else:
            return []

    def get_child(self, args, path=[]):
        name = args[0]
        args = args[1:]

        if self.filters.verbose > 1:
            print "[DeviceCollection.get_child {0} {1} {2}]".format(name, args, path)

        try:
            interface = self.device.get_interface(ifName=name)
        except InterfaceNotFound:
            return None

        path.append(name)
        if len(args):
            return interface.get_child(args, path=path)
        else:
            return interface

class Interface(NodeInfo):
    wrn = InterfaceWarning
    """Class to encapsulate interface information"""
    def __init__(self, data, api_url, filters):
        super(Interface, self).__init__(data, api_url, filters)

    # Property attrs unique to interfaces
    @property
    def device(self):
        return self._data.get('device', None)

    @property
    def ifAdminStatus(self):
        return self._data.get('ifAdminStatus', None)

    @property
    def ifAlias(self):
        return self._data.get('ifAlias', None)

    @property
    def ifDescr(self):
        return self._data.get('ifDescr', None)

    @property
    def ifName(self):
        return self._data.get('ifName', None)

    @property
    def ifHighSpeed(self):
        return self._data.get('ifHighSpeed', None)

    @property
    def ifIndex(self):
        return self._data.get('ifIndex', None)

    @property
    def ifMtu(self):
        return self._data.get('ifMtu', None)

    @property
    def ifOperStatus(self):
        return self._data.get('ifOperStatus', None)

    @property
    def ifPhysAddress(self):
        return self._data.get('ifPhysAddress', None)

    @property
    def ifSpeed(self):
        return self._data.get('ifSpeed', None)

    @property
    def ifType(self):
        return self._data.get('ifType', None)

    @property
    def ipAddr(self):
        return self._data.get('ipAddr', None)

    @property
    def uri(self):
        return self._data.get('uri', None)

    def get_endpoints(self):
        """
        Generate and return a list of endpoints (via a generator) associated
        with this interface.  Even though this does not 'do work' by issuing
        a call to the API, it is a get_ method just for consistency.
        """
        for i in self._data['children']:
            yield Endpoint(i, self.api_url, self.filters)

    def get_endpoint(self, name):
        endpoints = filter(lambda x: x.name == name, self.get_endpoints())

        if len(endpoints) == 0:
            return EndpointNotFound()
        elif len(endpoints) > 1:
            self.warn('Multiple endpoints found')

        return endpoints[0]

    def get_child(self, args, path=[]):
        name = args[0]
        args = args[1:]

        if self.filters.verbose > 1:
            print "[Interface.get_child {0} {1} {2}]".format(name, args, path)

        try:
            endpoint = self.get_endpoint(name)
        except EndpointNotFound:
            return None

        endpoint._data['leaf'] = True
        return endpoint

    def __repr__(self):
        return '<Interface/{0}: uri:{1}>'.format(self.ifName, self.resource_uri)

class Endpoint(NodeInfo):
    wrn = EndpointWarning
    """Class to encapsulate endpoint information"""
    def __init__(self, data, api_url, filters):
        super(Endpoint, self).__init__(data, api_url, filters)

    @property
    def name(self):
        return self._data.get('name', None)

    @property
    def uri(self):
        return self._data.get('uri', None)

    def get_data(self, **filters):
        """
        Retrieve the traffic data and return the json response in 
        a DataPayload object.  If there is something wrong with the 
        response (status/content) then issue a warning and return an 
        empty object.
        """

        r = requests.get('{0}{1}'.format(self.api_url, self.uri),
            params=self.filters.compose_filters(filters),
            headers=self.request_headers)

        self.inspect_request(r)

        if r.status_code == 200 and \
            r.headers['content-type'] == 'application/json':
            data = json.loads(r.text)
            return DataPayload(data)
        else:
            self.http_alert(r)
            return DataPayload()
        
    def __repr__(self):
        return '<Endpoint/{0}: uri:{1}>'.format(self.name, self.uri)

class DataPayload(NodeInfo):
    wrn = DataPayloadWarning
    """Class to encapsulate data payload"""
    def __init__(self, data={'data':[]}):
        super(DataPayload, self).__init__(data, None, None)

    @property
    def agg(self):
        return self._data.get('agg', None)

    @property
    def cf(self):
        return self._data.get('cf', None)

    @property
    def data(self):
        """Return internal data from payload as list of DataPoint."""
        return [DataPoint(**x) for x in self._data.get('data', [])]

    @property
    def dump(self):
        return self.pp.pformat(self.data)

    def __repr__(self):
        return '<DataPayload: len:{0} b:{1} e:{2}>'.format(
            len(self.data), self.begin_time, self.end_time)

class DataPoint(object):
    __slots__ = ['ts', 'val', 'm_ts']
    """Class to encapsulate the returned data points."""
    def __init__(self, ts, val, m_ts=None):
        super(DataPoint, self).__init__()
        self.ts = datetime.datetime.utcfromtimestamp(ts)
        self.val = val
        self.m_ts = m_ts

    @property
    def ts_epoch(self):
        return calendar.timegm(self.ts.utctimetuple())

    def __repr__(self):
        return '<DataPoint: ts:{0} val:{1}>'.format(self.ts, self.val)

class BulkDataPayload(DataPayload):
    wrn = BulkDataPayloadWarning
    """Class to encapsulate bulk data payload"""
    def __init__(self, data={'data':[]}):
        super(BulkDataPayload, self).__init__(data)

    @property
    def data(self):
        return [BulkDataRow(x) for x in self._data.get('data', [])]

    @property
    def device_names(self):
        return self._data.get('device_names', [])

    def __repr__(self):
        return '<BulkDataPayload: len:{0} devs:{1} b:{2} e:{3}>'.format(
            len(self._data.get('data', [])), len(self.device_names), self.begin_time, self.end_time)

class BulkDataRow(object):
    """Class to encapsulate and differentiate data from different devices,
    interfaces and direction/traffic endpoints when getting data back from 
    a bulk request for interface data."""
    def __init__(self, row={}):
        super(BulkDataRow, self).__init__()
        self._info = row.get('path', {})
        self._data = row.get('data', [])

    @property
    def device(self):
        return self._info.get('dev', None)

    @property
    def interface(self):
        return self._info.get('iface', None)

    @property
    def endpoint(self):
        return self._info.get('endpoint', None)

    @property
    def data(self):
        return [DataPoint(**x) for x in self._data]

    def __repr__(self):
        return '<BulkDataRow: dev:{0} iface:{1} endpoint:{2} len:{3}>'.format(
            self.device, self.interface, self.endpoint,len(self._data))

# - Query entry point and filtering.

class ApiFilters(AlertMixin, object):
    wrn = ApiFiltersWarning
    """Class to hold filtering/query options.  This will be used by 
    ApiConnect and also passed to all the encapsulation objects."""
    def __init__(self):
        super(ApiFilters, self).__init__()

        self._begin_time = datetime.datetime.utcfromtimestamp(int(time.time() - 3600))
        self._end_time = datetime.datetime.utcfromtimestamp(int(time.time()))

        # Values to use in GET requests - they are combined with a 
        # user defined dict of filtering args (agg, cf) and django
        # filtering options and passed as args to the GET query.
        self._default_filters = {
            'limit': 0,
        }

        # Values to use in POST requests, verbose flag, etc.
        self.verbose = False
        self.cf = 'average'
        self.agg = None
        # This needs to be checked via property.
        self._endpoint = ['in']

        self.auth_username = ''
        self.auth_apikey = ''

    def ts_epoch(self, time_prop):
        """Convert named property back to epoch.  Generally just for 
        sending cgi params to the api."""
        return int(calendar.timegm(getattr(self, time_prop).utctimetuple()))

    def _convert_ts(self, ts):
        """
        Massage the timestamp to datetime if need be.
        """
        if isinstance(ts, type(self._begin_time)):
            return ts
        else:
            return datetime.datetime.utcfromtimestamp(int(ts))

    def begin_time():
        doc = "The begin_time property."
        def fget(self):
            return self._begin_time
        def fset(self, value):
            self._begin_time = self._convert_ts(value)
            self._default_filters['begin'] = self.ts_epoch('begin_time')
        def fdel(self):
            pass
        return locals()
    begin_time = property(**begin_time())

    def end_time():
        doc = "The end_time property."
        def fget(self):
            return self._end_time
        def fset(self, value):
            self._end_time = self._convert_ts(value)
            self._default_filters['end'] = self.ts_epoch('end_time')
        def fdel(self):
            pass
        return locals()
    end_time = property(**end_time())

    def limit():
        doc = "The limit property."
        def fget(self):
            return self._limit
        def fset(self, value):
            self._limit = int(value)
            self._default_filters['limit'] = self._limit
        def fdel(self):
            del self._limit
        return locals()
    limit = property(**limit())

    def default_filters():
        doc = "The default_filters property."
        def fget(self):
            return self._default_filters
        def fset(self, value):
            self._default_filters = value
        def fdel(self):
            del self._default_filters
        return locals()
    default_filters = property(**default_filters())

    def endpoint():
        doc = "The endpoint property."
        def fget(self):
            return self._endpoint
        def fset(self, value):
            if not value or not isinstance(value, list):
                self.warn('The endpoint filter must be set to a non-empty list ("{0}" given), retaining current values {1}'.format(value,self.endpoint))
                return
            self._endpoint = value
        def fdel(self):
            del self._endpoint
        return locals()
    endpoint = property(**endpoint())

    def compose_filters(self, filters):
        """Compose filters using the defaults combined with local filters.

        The defaults will be overwritten by filters with the same name in the
        local filters."""

        return dict(self.default_filters, **filters)


class ApiConnect(AlertMixin, object):
    wrn = ApiConnectWarning
    """Core class to pull data from the rest api."""
    def __init__(self, api_url, filters=ApiFilters(), username='', api_key=''):
        super(ApiConnect, self).__init__()
        self.api_url = api_url.rstrip("/")
        self.filters = filters
        self.filters.auth_username = username
        self.filters.auth_apikey = api_key
        self._valid_endpoints = []

        self.request_headers = {}
        if self.filters.auth_username and self.filters.auth_apikey:
            add_apikey_header(self.filters.auth_username, 
                self.filters.auth_apikey, self.request_headers)

    def get_devices(self, **filters):
        r = requests.get('{0}/v1/device/'.format(self.api_url),
            params=self.filters.compose_filters(filters),
            headers=self.request_headers)

        self.inspect_request(r)
        
        if r.status_code == 200 and \
            r.headers['content-type'] == 'application/json':
            data = json.loads(r.text)
            for i in data:
                yield Device(i, self.api_url, self.filters)
        else:
            self.http_alert(r)
            return
            yield

    def get_device(self, name):
        devices = list(self.get_devices(name=name))

        if len(devices) == 0:
            raise DeviceNotFound()

        return devices[0]
        
    @property
    def children(self):
        return [ device.name for device in self.get_devices() ]

    def get_child(self, args, path=[]):
        name = args[0]
        args = args[1:]

        if self.filters.verbose > 1:
            print "[ApiConnect.get_child {0} {1} {2}]".format(name, args, path)

        try:
            device = self.get_device(name)
        except DeviceNotFound:
            return None

        path.append(name)
        if args:
            return device.get_child(args, path=path)
        else:
            return device

    def get_interfaces(self, **filters):
        r = requests.get('{0}/v1/interface/'.format(self.api_url),
                params=self.filters.compose_filters(filters),
                headers=self.request_headers)

        self.inspect_request(r)

        if r.status_code == 200 and \
            r.headers['content-type'] == 'application/json':
            data = json.loads(r.text)
            for i in data['children']:
                yield Interface(i, self.api_url, self.filters)
        else:
            self.http_alert(r)
            return
            yield

    def get_interface_bulk_data(self, **filters):
        interfaces = []

        for i in self.get_interfaces(**filters):
            if self.filters.verbose > 1: print i
            interfaces.append({'device': i.device, 'iface': i.ifName})
            # if self.filters.verbose > 1: print i.dump

        return self._execute_get_interface_bulk_data(interfaces)

    def _execute_get_interface_bulk_data(self, interfaces=[]):

        if not self._check_endpoints():
            return BulkDataPayload()

        payload = { 
            'interfaces': interfaces, 
            'endpoint': self.filters.endpoint,
            'cf': self.filters.cf,
            'begin': self.filters.ts_epoch('begin_time'),
            'end': self.filters.ts_epoch('end_time'),
        }

        if self.filters.agg: payload['agg'] = self.filters.agg

        self.request_headers['content-type'] = 'application/json'

        r = requests.post('{0}/v1/bulk/interface/'.format(self.api_url), 
            headers=self.request_headers, data=json.dumps(payload))

        self.inspect_request(r)
        self.inspect_payload(payload)

        if r.status_code == 201 and \
            r.headers['content-type'] == 'application/json':
            return BulkDataPayload(json.loads(r.text))
        else:
            self.http_alert(r)
            return BulkDataPayload()

    def _check_endpoints(self):
        if not self._valid_endpoints:
            r = requests.get('{0}/v1/oidsetmap/'.format(self.api_url),
                headers=self.request_headers)
            if not r.status_code == 200:
                self.warn('Could not retrieve oid set map from REST api.')
                return False
            data = json.loads(r.content)
            for i in data.keys():
                for ii in data[i].keys():
                    if ii not in self._valid_endpoints:
                        self._valid_endpoints.append(ii)

        for ep in self.filters.endpoint:
            if ep not in self._valid_endpoints:
                self.warn('{0} is not a valid endpoint type - must be of the form {1} - cancelling query'.format(ep, self._valid_endpoints))
                return False

        return True

    def inspect_request(self, r):
        if self.filters.verbose:
            print '[url: {0}]'.format(r.url)

    def inspect_payload(self, p):
        if self.filters.verbose > 1:
            print '[POST payload: {0}]'.format(json.dumps(p, indent=4))









#!/usr/bin/env python

import os
from setuptools import setup

def read(*paths):
    """Build a file path from *paths* and return the contents."""
    with open(os.path.join(*paths), 'r') as f:
        return f.read()

setup(
    name='esmond_client',
    version='1.3',
    description='API client libraries and command line tools for the ESnet Monitoring Daemon (esmond).',
    long_description=(read('README.rst')),
    author='Monte M. Goode',
    author_email='MMGoode@lbl.gov',
    url='http://software.es.net/esmond/',
    packages=['esmond_client', 'esmond_client.perfsonar'],
    scripts=[
        'clients/esmond-ps-get',
        'clients/esmond-ps-get-bulk',
        'clients/esmond-ps-get-endpoints', 
        'clients/esmond-ps-get-metadata',
    ],
    install_requires=['requests', 'python-dateutil'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: Telecommunications Industry',
        'Environment :: Console',
        'License :: OSI Approved :: BSD License',
        'Operating System :: POSIX',
        'Programming Language :: Python :: 2',
        'Topic :: Internet',
        'Topic :: System :: Networking',
        'Topic :: Software Development :: Libraries',
    ],
)



import calendar
import os.path
import sys
import warnings

import ConfigParser

"""Utils for esmond.api.client modules and scripts."""

def add_apikey_header(user, key, header_dict):
    """Format an auth header for rest api key."""
    header_dict['Authorization'] = 'ApiKey {0}:{1}'.format(user, key)

class AlertMixin(object):
    def http_alert(self, r):
        """
        Issue a subclass specific alert in the case that a call to the REST
        api does not return a 200 status code.
        """
        warnings.warn('Request for {0} got status: {1} - response: {2}'.format(r.url,r.status_code,r.content), self.wrn, stacklevel=2)

    def warn(self, m):
        warnings.warn(m, self.wrn, stacklevel=2)

# -- defines and config handling for summary scripts/code

SUMMARY_NS = 'summary'
MONTHLY_NS = 'monthly'

class ConfigException(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)

class ConfigWarning(Warning): pass

def get_config():
    c_path = os.path.abspath(sys.argv[0])
    if c_path.endswith('.py'):
        c_path = c_path.replace('.py', '.conf')
    else:
        c_path = c_path + '.conf'
    if not os.path.exists(c_path):
        raise ConfigException('Could not find configuration file {0}'.format(c_path))

    config = ConfigParser.ConfigParser()
    config.read(c_path)
    return config

def get_type_map():
    type_map = {}

    c = get_config()
    for section in c.sections():
        type_map[section] = {}
        for items in c.items(section):
            type_map[section][items[0]] = items[1]

    return type_map

def get_summary_name(filterdict):
    if not isinstance(filterdict, dict):
        raise ConfigException('Arg needs to be a dict of the form: {{django_query_filter: filter_criteria}} - got {0}.'.format(filterdict))
    elif len(filterdict.keys()) > 1:
        raise ConfigException('Dict must contain a single key/value pair of the form: {{django_query_filter: filter_criteria}} - got {0}.'.format(filterdict))

    django_query_filter = filterdict.keys()[0]
    filter_criteria = filterdict[django_query_filter]

    type_map = get_type_map()

    if not type_map.has_key(django_query_filter):
        raise ConfigException('Config file did does not contain a section for {0} - has: {1}'.format(django_query_filter, type_map.keys()))
    elif not type_map[django_query_filter].has_key(filter_criteria):
        raise ConfigException('Config section for {0} does not contain an key/entry for {1} - has: {2}'.format(django_query_filter, filter_criteria, type_map[django_query_filter].keys()))

    return type_map[django_query_filter][filter_criteria]

# -- aggregation functions

def aggregate_to_ts_and_endpoint(data, verbosity=False):
    aggs = {}

    # Aggregate/sum the returned data by timestamp and endpoint alias.
    for row in data.data:
        if verbosity: print ' *', row
        for data in row.data:
            if verbosity > 1: print '  *', data
            if not aggs.has_key(data.ts_epoch): aggs[data.ts_epoch] = {}
            if not aggs[data.ts_epoch].has_key(row.endpoint): 
                aggs[data.ts_epoch][row.endpoint] = 0
            if data.val != None:
                aggs[data.ts_epoch][row.endpoint] += data.val
        pass

    return aggs

def aggregate_to_device_interface_endpoint(data, verbosity=False):
    aggs = {}

    # Aggregate/sum the returned data to device/interface/endpoint alias.
    for row in data.data:
        if verbosity: print ' *', row
        if not aggs.has_key(row.device): aggs[row.device] = {}
        if not aggs[row.device].has_key(row.interface): 
            aggs[row.device][row.interface] = {}
        if not aggs[row.device][row.interface].has_key(row.endpoint): 
            aggs[row.device][row.interface][row.endpoint] = 0
        for data in row.data:
            if verbosity > 1: print '  *', data
            if data.val != None:
                aggs[row.device][row.interface][row.endpoint] += data.val

    return aggs

def iterate_device_interface_endpoint(aggs, verbosity=False):
    for device in aggs.keys():
        for interface in aggs[device].keys():
            for endpoint,val in aggs[device][interface].items():
                yield device, interface, endpoint, val


# -- timehandling code for summary scripts

lastmonth = lambda (yr,mo): [(y,m+1) for y,m in (divmod((yr*12+mo-2), 12),)][0]
nextmonth = lambda (yr,mo): (yr+mo/12,mo%12+1)

def get_month_start_and_end(start_point):

    start = calendar.timegm((start_point.year, start_point.month, 1, 0,0,0,0,0,-1))

    n_mo_yr, n_mo = nextmonth((start_point.year, start_point.month))

    end = calendar.timegm((n_mo_yr, n_mo, 1, 0,0,0,0,0,-1)) - 1

    return start, end


# -- atencode code for handling rest URIs

_atencode_safe = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXWZ0123456789_.-'
_atencode_map = {}
for i, c in (zip(xrange(256), str(bytearray(xrange(256))))):
    _atencode_map[c] = c if c in _atencode_safe else '@{:02X}'.format(i)

_atencode_unsafe = ' $&+,/:;=?@\x7F'
_atencode_map_minimal = {}
for i, c in (zip(xrange(256), str(bytearray(xrange(256))))):
    _atencode_map_minimal[c] = c if (i > 31 and i < 128 and c not in _atencode_unsafe) else '@{:02X}'.format(i)

_atencode_safe_graphite = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXWZ012345689_-'
_atencode_map_graphite = {}
for i, c in (zip(xrange(256), str(bytearray(xrange(256))))):
    _atencode_map_graphite[c] = c if c in _atencode_safe_graphite else '@{:02X}'.format(i)

_atdecode_map = {}

for i in xrange(256):
    _atdecode_map['{:02X}'.format(i)] = chr(i)
    _atdecode_map['{:02x}'.format(i)] = chr(i)

def atencode(s, minimal=False, graphite=False):
    if minimal:
        return ''.join(map(_atencode_map_minimal.__getitem__, s))
    elif graphite:
        return ''.join(map(_atencode_map_graphite.__getitem__, s))
    else:
        return ''.join(map(_atencode_map.__getitem__, s))

def atdecode(s):
    parts = s.split('@')
    r = [parts[0]]

    for part in parts[1:]:
        try:
            r.append(_atdecode_map[part[:2]])
            r.append(part[2:])
        except KeyError:
            append('@')
            append(part)

    return ''.join(r)





"""
Client to fetch cassandra/os information from MX4J

See:

http://mx4j.sourceforge.net/
http://wiki.apache.org/cassandra/Operations#Monitoring_with_MX4J
"""

import urllib
import warnings
import xml.etree.ElementTree as ET

import requests

class CassandraJMXException(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)

class CassandraJMXWarning(Warning): pass

class CassandraJMX(object):
    def __init__(self, url='http://localhost:8081'):
        self.url = url.rstrip('/')
        # - JMX variables
        # Activity/number of ops
        self.jmx_readstage = 'org.apache.cassandra.request:type=ReadStage'
        self.jmx_mutationstage = 'org.apache.cassandra.request:type=MutationStage'
        self.jmx_gossipstage = 'org.apache.cassandra.internal:type=GossipStage'
        # Latency
        self.jmx_storageproxy = 'org.apache.cassandra.db:type=StorageProxy'
        # Memory
        self.jmx_memory = 'java.lang:type=Memory'
        # Java garbage collection
        self.jmx_garbage = 'java.lang:type=GarbageCollector,name=ConcurrentMarkSweep'
        # Operating system information
        self.jmx_os = 'java.lang:type=OperatingSystem'
        # Compaction information
        self.jmx_compaction = 'org.apache.cassandra.db:type=CompactionManager'

    def _make_request(self, var):
        qs = dict(objectname=var, template='identity')
        url = '{0}/mbean?{1}'.format(self.url, urllib.urlencode(qs))
        r = requests.get(url)
        if r.status_code != 200:
            warnings.warn('Bad request: {0} got return code: {1}'.format(url, r.status_code), 
                CassandraJMXWarning, stacklevel=2)
        return r.content

    def _get_attribute_value(self, root, name):
        for i in root.iterfind('Attribute[@name="{0}"]'.format(name)):
            return i.attrib['value']

    def _fetch_value(self, jmx_var, attr):
        root = ET.fromstring(self._make_request(jmx_var))
        val = self._get_attribute_value(root, attr)
        try:
            val = int(val)
        except ValueError:
            pass
        return val

    def _get_contents_dict(self, s):
        d = {}
        for i in s[s.find('contents=')+10:-2].split(','):
            k,v = i.strip().split('=')
            d[k] = int(v)
        return d

    # Memory stats

    def get_heap_memory(self):
        value = self._fetch_value(self.jmx_memory, 'HeapMemoryUsage')
        return self._get_contents_dict(value)

    def get_non_heap_memory(self):
        value = self._fetch_value(self.jmx_memory, 'NonHeapMemoryUsage')
        return self._get_contents_dict(value)

    # Latency stats

    def get_read_latency(self):
        return self._fetch_value(self.jmx_storageproxy, 'RecentReadLatencyMicros')

    def get_write_latency(self):
        return self._fetch_value(self.jmx_storageproxy, 'RecentWriteLatencyMicros')

    def get_range_latency(self):
        return self._fetch_value(self.jmx_storageproxy, 'RecentRangeLatencyMicros')

    # Java garbage collector stats

    def get_gc_count(self):
        return self._fetch_value(self.jmx_garbage, 'CollectionCount')

    def get_gc_time(self):
        return self._fetch_value(self.jmx_garbage, 'CollectionTime')

    # Read/write/gossip operations

    def get_read_active(self):
        return self._fetch_value(self.jmx_readstage, 'ActiveCount')

    def get_read_pending(self):
        return self._fetch_value(self.jmx_readstage, 'PendingTasks')

    def get_read_completed(self):
        return self._fetch_value(self.jmx_readstage, 'CompletedTasks')

    def get_write_active(self):
        return self._fetch_value(self.jmx_mutationstage, 'ActiveCount')

    def get_write_pending(self):
        return self._fetch_value(self.jmx_mutationstage, 'PendingTasks')

    def get_write_completed(self):
        return self._fetch_value(self.jmx_mutationstage, 'CompletedTasks')

    def get_gossip_active(self):
        return self._fetch_value(self.jmx_gossipstage, 'ActiveCount')

    def get_gossip_pending(self):
        return self._fetch_value(self.jmx_gossipstage, 'PendingTasks')

    def get_gossip_completed(self):
        return self._fetch_value(self.jmx_gossipstage, 'CompletedTasks')

    # OS information

    def get_os_load(self):
        return self._fetch_value(self.jmx_os, 'SystemLoadAverage')

    def get_os_free_memory(self):
        return self._fetch_value(self.jmx_os, 'FreePhysicalMemorySize')

    def get_os_free_swap(self):
        return self._fetch_value(self.jmx_os, 'FreeSwapSpaceSize')

    def get_os_committed_virtual_memory(self):
        return self._fetch_value(self.jmx_os, 'CommittedVirtualMemorySize')

    # Compaction information

    def get_compaction_pending(self):
        return self._fetch_value(self.jmx_compaction, 'PendingTasks')

    def get_compaction_complete(self):
        return self._fetch_value(self.jmx_compaction, 'CompletedTasks')

available_tests = filter(lambda x: x.startswith('get_'), dir(CassandraJMX('')))
